//
//  Core.cpp
//  Ice
//  This class file is the core of the simulation.


#include "Core.hpp"
double cube::mass()
{
	double mass = 0;
	std::map<int,double>::iterator it;
	for(it = masses.begin(); it != masses.end(); it++)
		{
		std::cout<<(*it).first<<" "<<(*it).second<<std::endl;
		mass+=(*it).second;
		}
	return mass;
}
double cube::charge()
{
	return 0;
}
void cube::addDust(int dustID, double mass)
{
	masses[dustID] += mass;
}
void cube::removeDust(int dustID, double mass)
{
	masses[dustID] -= mass;
}
automaton::automaton(vector loc)
{
	location.x = loc.x;
	location.y = loc.y;
	location.z = loc.z;
}
automaton::~automaton() //For the worker thread later
{
	join();
}
void automaton::start()
{
	m_thread = boost::thread(&automaton::callUponKronos, this, location);
}
void automaton::join()
{
	m_thread.join();
}
bool automaton::timedJoin(int seconds)
{
	using namespace boost::interprocess;
	try{
	return m_thread.timed_join(boost::posix_time::milliseconds(seconds));
	}
	catch(interprocess_exception &ex){
		std::cout << ex.what() << std::endl;
	}
}
void automaton::callUponKronos(vector loc)
//Tells the worker thread to begin its calculations
//This is where the majority of our calculations are.
{
	//sleep(20);

	location.x = loc.x;
	location.y = loc.y;
	location.z = loc.z;

	using namespace boost::interprocess;
	
	shared_memory_object shm1(open_only, memname, read_write);
    mapped_region region1(shm1, read_write);
	dataplexor *plex = static_cast<dataplexor*>(region1.get_address());
	
	sharable_lock<interprocess_upgradable_mutex> *lock = new sharable_lock<interprocess_upgradable_mutex>(plex->mutex);

	std::map<vector,cube>::iterator it;
	cube *data = &plex->cubeArray[location];
	vector direction(data->direction.x,data->direction.y,data->direction.z);
	double p, dist, sigma, phi;
    for(it = plex->cubeArray.begin(); it != plex->cubeArray.end(); it++)
    {
    	if((*it).first != location)
    	{	
		//Gravitational Simluation
		//The Math is here
		//The Distance Between Points = (x^2 + y^2 + z^2)^.5
		p = plex->cubeSize*pow(pow((*it).second.location.x-data->location.x,2)+pow((*it).second.location.z-data->location.z,2)+pow((*it).second.location.y-data->location.y,2),.5);
		
		//The Distance Moved
		dist = pow(time,2)*(gconst*(*it).second.mass())/(2*pow(p,2));
		
		//Determine Sigma Angle
			sigma = atan(((*it).second.location.y - data->location.y)/pow(pow((*it).second.location.z - data->location.z,2) + pow((*it).second.location.x - data->location.x,2),.5));
		//Determine Phi Angle
			phi = atan(((*it).second.location.z - data->location.z)/((*it).second.location.x - data->location.x));
		
		//Vector Addition
			direction.x += cos(phi)*cos(sigma)*dist;
			direction.y += sin(sigma)*dist;
			direction.z += sin(phi)*cos(sigma)*dist;
		}
    }
	
    shared_memory_object shm(open_only, memcopyname, read_write);
    mapped_region region(shm, read_write);
	dataplexor *cdata = static_cast<dataplexor*>(region.get_address());
	scoped_lock<interprocess_upgradable_mutex> clock(cdata->mutex);
		
	cube *newCube = cdata->cubeAt(location);
	double *eight = (double*)calloc(8,sizeof(double));
	
	float temp;
	
	//0 = Top-Bottom-Left, 4 = Top-Top-Left, 1 = Top-Bottom-Right, 5 = Top-Top-Right, 2 = Bottom-Bottom-Left, ...
    
	//Split Along X-Axis
	if(abs(direction.x - floor(direction.x)) < .5)
	   {
	   temp = abs(direction.x - floor(direction.x));
	   eight[0] = .5 - temp;
	   eight[4] = .5 - temp;
	   eight[3] = .5 - temp;
	   eight[6] = .5 - temp;
	   eight[1] = .5 + temp;
	   eight[5] = .5 + temp;
	   eight[3] = .5 + temp;
	   eight[7] = .5 + temp;
	   }
	else
	   {
	   temp = abs(floor(direction.x+1) - direction.x);
	   eight[0] = .5 + temp;
	   eight[4] = .5 + temp;
	   eight[3] = .5 + temp;
	   eight[6] = .5 + temp;
	   eight[1] = .5 - temp;
	   eight[5] = .5 - temp;
	   eight[3] = .5 - temp;
	   eight[7] = .5 - temp;
	   }
	//Split Along Z-Axis
	if(abs(direction.z - floor(direction.z)) < .5)
	   {
	   temp = abs(direction.z - floor(direction.z));
	   eight[0] = eight[0]*(.5 - temp);
	   eight[4] = eight[4]*(.5 - temp);
	   eight[1] = eight[1]*(.5 - temp);
	   eight[5] = eight[5]*(.5 - temp);
	   eight[2] = eight[2]*(.5 + temp);
	   eight[6] = eight[6]*(.5 + temp);
	   eight[3] = eight[3]*(.5 + temp);
	   eight[7] = eight[7]*(.5 + temp);
	   }
	else
	   {
	   temp = abs(floor(direction.z+1) - direction.z);
	   eight[0] = eight[0]*(.5 + temp);
	   eight[4] = eight[4]*(.5 + temp);
	   eight[1] = eight[1]*(.5 + temp);
	   eight[5] = eight[5]*(.5 + temp);
	   eight[2] = eight[2]*(.5 - temp);
	   eight[6] = eight[6]*(.5 - temp);
	   eight[3] = eight[3]*(.5 - temp);
	   eight[7] = eight[7]*(.5 - temp);
	   }
	//Split Along Y-Axis
	if(abs(direction.y - floor(direction.y)) < .5)
	   {
	   temp = abs(direction.y - floor(direction.y));
	   eight[4] = eight[4]*(.5 + temp);
	   eight[4] = eight[5]*(.5 + temp);
	   eight[6] = eight[6]*(.5 + temp);
	   eight[7] = eight[7]*(.5 + temp);
	   eight[0] = eight[0]*(.5 - temp);
	   eight[1] = eight[1]*(.5 - temp);
	   eight[2] = eight[2]*(.5 - temp);
	   eight[3] = eight[3]*(.5 - temp);
	   }
	   else
	   {
	   temp = abs(floor(direction.y+1) - direction.y);
	   eight[4] = eight[4]*(.5 - temp);
	   eight[5] = eight[5]*(.5 - temp);
	   eight[6] = eight[6]*(.5 - temp);
	   eight[7] = eight[7]*(.5 - temp);
	   eight[0] = eight[0]*(.5 + temp);
	   eight[1] = eight[1]*(.5 + temp);
	   eight[2] = eight[2]*(.5 + temp);
	   eight[3] = eight[3]*(.5 + temp);
	   }
	//Here we divide mass among the cubes...
	
	//vector(floor(direction.x-.5)+.5),(floor(direction.y-.5)+.5),(floor(direction.z+.5)+.5));
	std::map<int,double>::iterator it1;
	for(it1 = data->masses.begin(); it1 != data->masses.end(); it1++)
		{
		cdata->cubeAt((floor(direction.x-.5)+.5),(floor(direction.y-.5)+.5),(floor(direction.z+.5)+.5))->addDust((*it1).first,eight[0]*(*it1).second);//Cube 1
	cdata->cubeAt((floor(direction.x+.5)+.5),(floor(direction.y-.5)+.5),(floor(direction.z+.5)+.5))->addDust((*it1).first,eight[1]*(*it1).second);//Cube 2
	cdata->cubeAt((floor(direction.x-.5)+.5),(floor(direction.y-.5)+.5),(floor(direction.z-.5)+.5))->addDust((*it1).first,eight[2]*(*it1).second);//Cube 3
	cdata->cubeAt((floor(direction.x+.5)+.5),(floor(direction.y-.5)+.5),(floor(direction.z-.5)+.5))->addDust((*it1).first,eight[3]*(*it1).second);//Cube 4
	cdata->cubeAt((floor(direction.x-.5)+.5),(floor(direction.y+.5)+.5),(floor(direction.z+.5)+.5))->addDust((*it1).first,eight[4]*(*it1).second);//Cube 5
	cdata->cubeAt((floor(direction.x+.5)+.5),(floor(direction.y+.5)+.5),(floor(direction.z+.5)+.5))->addDust((*it1).first,eight[5]*(*it1).second);//Cube 6
	cdata->cubeAt((floor(direction.x-.5)+.5),(floor(direction.y+.5)+.5),(floor(direction.z-.5)+.5))->addDust((*it1).first,eight[6]*(*it1).second);//Cube 7
	cdata->cubeAt((floor(direction.x+.5)+.5),(floor(direction.y+.5)+.5),(floor(direction.z-.5)+.5))->addDust((*it1).first,eight[7]*(*it1).second);//Cube 8
		}
	//	newCube
	//data 
	clock.unlock();
	lock->unlock();
}
cube* dataplexor::cubeAt(vector loc)
{
	//if (cubeArray.count(loc) == 0)
		//cubeArray[loc] = cube();
	return &cubeArray[loc];
}
cube* dataplexor::cubeAt(int x, int y, int z)
{
	return &cubeArray[vector(x,y,z)];
}
void initiateFusion()
{
   //Wait for Scenario to be loaded
    bool stop = false;
    while(stop != true)
    {
	using namespace boost::interprocess;
	try{
		//Read & Lock Scenario
    shared_memory_object shm(open_only, memname, read_write);
    mapped_region region(shm, read_write);
	dataplexor *data = static_cast<dataplexor*>(region.get_address());
	sharable_lock<interprocess_upgradable_mutex> lock(data->mutex);

	//Make Copy
	shared_memory_object::remove(memcopyname);
	shared_memory_object shmB(create_only, memcopyname, read_write);
	shmB.truncate(sizeof(*data));
	mapped_region regionB(shmB, read_write);
	dataplexor *copy = static_cast<dataplexor*>(regionB.get_address());
	memcpy(copy,data,sizeof(*data));
	
	//Clear Copy Data
	std::map<vector,cube>::iterator it;
	for (it = copy->cubeArray.begin(); it != copy->cubeArray.begin(); it++) {
		(*it).second.masses.clear();
	}		
		
	//Determine the number of physical threads available
		int nthreads;
		if(boost::thread::hardware_concurrency() != 0)
			nthreads = boost::thread::hardware_concurrency();
		else
			nthreads = 2;
		
		std::map<int, bool> completion;
		for (int u = 0; u < nthreads; u++)
			completion[u] = false;
		int complete = 0;
		
		//
		std::map<vector,cube>::const_iterator it2;
		std::map<int,automaton*> threads;
		
		int i = 0;
		for(it2 = data->cubeArray.begin(); it2 != data->cubeArray.begin(); it2++)
			{
			if(i < nthreads)
				{
				//Thread Check
				threads[i] = new automaton(it2->first);
				threads[i]->start();
				i++;
				}
			else
				it2 = data->cubeArray.begin();
			}
		//
		while (complete != nthreads) {
			for(int	i = 0; i < nthreads; i++)
				{
				if (threads[i]->timedJoin(500)) 
				{
					if(it != copy->cubeArray.end() && !completion[i])
						{
						delete(threads[i]);
						threads[i] = new automaton((*it2).first);
						threads[i]->start();
						it2++;
						}
					else
						{
						completion[i] = true;
						complete += 1;
						}
				}
			}
		}
    //Check to see if pause is requested...
	if(data->halt == true)
	{
	    stop = true;
        data->halt = false;
	}
	//Unlock Memory
	lock.unlock();
	}
	//Catch Exceptions
	catch(interprocess_exception &ex){
		shared_memory_object::remove(memname);
		std::cout << ex.what() << std::endl;
	}
    }
}